/*
** Module   : SOAPFaultImpl.java
** Abstract : Implementation for the SOAP-FAULT resource.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130216 Created initial version.
** 002 CA  20130221 Added Deletable support. Fixed NAME and TYPE support.
** 003 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 004 CA  20130927 Removed setResourceType, as the type is determined from LegacyResource 
**                  annotation.
** 005 CA  20140104 Added runtime implementation for this resource.
*/

package com.goldencode.p2j.util;

/**
 * A class for implementing the SOAP-FAULT method and attributes.
 */
public class SOAPFaultImpl
extends HandleResource
implements SOAPFault,
           Deletable
{
   /** The URI of the Web service actor that caused this SOAP-FAULT. */
   private final String faultActor;
   
   /** The SOAP-FAULT-CODE of this SOAP-FAULT object. */
   private final String faultCode;
   
   /** The SOAP-FAULT-STRING of this SOAP-FAULT object. */
   private final String faultString;
   
   /** The SOAP-FAULT-DETAIL associated with this SOAP-FAULT object. */
   private final SOAPFaultDetailImpl faultDetail;
   
   /** Flag indicating this resource was deleted or not. */
   private boolean deleted = false;
   
   /**
    * Create a new SOAP Fault instance.
    * 
    * @param    faultActor
    *           The URI of the Web service actor that caused this SOAP-FAULT.
    * @param    faultCode
    *           The SOAP-FAULT-CODE of this SOAP-FAULT object.
    * @param    faultString
    *           The SOAP-FAULT-STRING of this SOAP-FAULT object.
    * @param    faultDetail
    *           The SOAP-FAULT-STRING of this SOAP-FAULT object.
    */
   SOAPFaultImpl(String              faultActor,
                 String              faultCode,
                 String              faultString,
                 SOAPFaultDetailImpl faultDetail)
   {
      this.faultActor = faultActor;
      this.faultCode = faultCode;
      this.faultString = faultString;
      this.faultDetail = faultDetail;
   }
   
   /**
    * The URI of the Web service actor that caused this SOAP-FAULT.
    * 
    * @return   See above.
    */
   public character getFaultActor()
   {
      return new character(faultActor);
   }

   /**
    * The SOAP-FAULT-CODE of this SOAP-FAULT object.
    * 
    * @return   See above.
    */
   public character getFaultCode()
   {
      return new character(faultCode);
   }

   /**
    * Returns the handle of the SOAP-FAULT-DETAIL associated with this SOAP-FAULT object.
    * 
    * @return   See above.
    */
   public handle getFaultDetail()
   {
      return new handle(faultDetail);
   }

   /**
    * The SOAP-FAULT-STRING of this SOAP-FAULT object.
    * 
    * @return   See above.
    */
   public character getFaultString()
   {
      return new character(faultString);
   }
   
   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return !deleted;
   }

   /**
    * Delete this resource. This just sets the {@link #deleted} flag to {@code true}.
    * <p>
    * The referenced {@link #faultDetail} will be deleted too.
    */
   @Override
   public void delete()
   {
      if (deleted)
      {
         return;
      }

      deleted = true;
      
      if (faultDetail != null)
      {
         // the fault detail is deleted if the parent soap fault is deleted.
         faultDetail.delete();
      }
   }
}
