/*
** Module   : SOAPHeaderEntryImpl.java
** Abstract : Implementation for the SOAP-HEADER-ENTRY resource.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130216 Created initial version.
** 002 CA  20130221 Added ADM-DATA, LOCAL-NAME, UNIQUE-ID and NAMESPACE-URI support. Fixed 
**                  resource type.
** 003 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 004 CA  20130603 resourceDelete must return a boolean value.
** 005 CA  20130927 Removed setResourceType, as the type is determined from LegacyResource 
**                  annotation.
** 006 HC  20131215 Implemented ADM-DATA and UNIQUE-ID attribute.
** 007 CA  20140107 Added runtime support.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.logging.*;

import javax.xml.parsers.*;

import org.w3c.dom.*;
import org.xml.sax.*;

import com.goldencode.p2j.xml.*;
import com.goldencode.util.*;
import com.goldencode.p2j.util.UniqueIdGenerator.IdKind;

/**
 * A class for implementing the SOAP-HEADER-ENTRYREF methods and attributes.
 */
public class SOAPHeaderEntryImpl
extends HandleChain
implements SOAPHeaderEntry,
           SOAPEntity
{
   /** Key identifying the ACTOR of this SOAP Header Entry, set at the DOM Node. */
   static final String ACTOR = "ACTOR";
   
   /**
    * Key identifying the MUST-UNDERSTAND attribute of this SOAP Header Entry, set at the DOM 
    * Node.
    */
   static final String MUST_UNDERSTAND = "MUST-UNDERSTAND";

   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(SOAPHeaderEntryImpl.class.getName());
   
   /** Corresponds to ADM-DATA attribute. */
   private final character admData = new character();
   
   /** Corresponds to UNIQUE-ID attribute. */
   private final integer uniqueID = new integer(UniqueIdGenerator.getUniqueId(IdKind.SOAP_HEADER));

   /** Flag indicating if this resource was deleted or not. */
   private boolean deleted = false;
   
   /**
    * After it is attached to a SOAP Header parent, this will be the key to the DOM Node for this
    * header entry.
    */
   private Object headerKey = null;
   
   /** The parent SOAP Header. */
   private SOAPHeaderImpl parent = null;
   
   /**
    * Implicit c'tor.
    */
   public SOAPHeaderEntryImpl()
   {
      super(true, false);
   }
   
   /**
    * Gets the X-NODEREF object that refers to the XML underlying a SOAP-HEADER-ENTRYREF object or
    * SOAP-FAULT entry and sets it to the given handle..
    * <p>
    * This method returns a logical even though the Progress 4GL documentation explicitly states 
    * otherwise. Actual working code has proven that the original documentation is incorrect. 
    * This implementation matches the actual runtime behavior of the 4GL.
    * 
    * @param    xNodeRef
    *           The handle where the resource will be set.
    *           
    * @return   <code>true</code> if the operation was possible.
    */
   public logical getNode(handle xNodeRef)
   {
      if (xNodeRef._isValid() && (xNodeRef.getResource() instanceof XNodeRefImpl))
      {
         Node node = getHeaderNode();
         
         if (node == null)
         {
            // if not attached, return false;
            return new logical(false);
         }
         
         XNodeRefImpl nodeRef = (XNodeRefImpl) xNodeRef.getResource();
         nodeRef.setNode(node);
         
         return new logical(true);
      }
      
      SOAPFactory.throwError("GET-NODE", "X-NODEREF");

      return new logical(false);
   }

   /**
    * Returns a LONGCHAR that contains the serialized form of the XML underlying the 
    * SOAP-HEADER-ENTRYREF or SOAP-FAULT entry.
    * 
    * @return   See above.
    */
   public longchar getSerialized()
   {
      longchar lc = new longchar();
      Node node = getHeaderNode();
      
      if (node == null)
      {
         // return unknown if node is not set
         return lc;
      }
      
      ByteArrayOutputStream output = new ByteArrayOutputStream();
      try
      {
         XmlHelper.write(node.getOwnerDocument(), output, false, true);
      }
      catch (IOException e)
      {
         throw new RuntimeException("Could not serialize SOAP Header Entry!", e);
      }

      lc.assign(output.toString());
      
      return lc;
   }

   /**
    * Returns the value of the actor attribute for the SOAP-HEADER-ENTRYREF object as a URL. 
    * Identifies the recipient of a header element.
    * 
    * @return   See above.
    */
   public character getActor()
   {
      Node node = getHeaderNode();

      if (node == null)
      {
         return new character("");
      }
      
      return (node == null || node.getUserData(ACTOR) == null)
                ? new character("") 
                : new character((character) node.getUserData(ACTOR)); 
   }

   /**
    * Indicates whether a SOAP-HEADER-ENTRYREF object is mandatory (TRUE) or optional (FALSE) for
    * the recipient to process.
    * 
    * @return   See above.
    */
   public logical isMustUnderstand()
   {
      Node node = getHeaderNode();
      
      if (node == null)
      {
         return new logical(false);
      }

      return new logical(node != null && Boolean.TRUE.equals(node.getUserData(MUST_UNDERSTAND)));
   }

   /**
    * Deletes the XML underlying a SOAP-HEADER-ENTRYREF object, without deleting the object.
    * 
    * @return   <code>true</code> if the operation was possible.
    */
   public logical deleteHeaderEntry()
   {
      if (parent == null)
      {
         return new logical(false);
      }

      // remove the node from the parent and detach the headerKey from all references
      parent.removeHeaderEntry(this, headerKey);
      
      // detach references
      parent = null;
      
      return new logical(true);
   }

   /**
    * Sets the ACTOR attribute for this SOAP-HEADER-ENTRY.
    *
    * @param    uri
    *           The URI of the SOAP actor.
    * 
    * @return   <code>true</code> if the operation was possible.
    */
   public logical setActor(character uri)
   {
      Node node = getHeaderNode();

      if (!isPopulated("SET-ACTOR"))
      {
         return new logical(false);
      }
      
      node.setUserData(ACTOR, new character(uri), null);
      
      return new logical(true);
   }

   /**
    * Sets the ACTOR attribute for this SOAP-HEADER-ENTRY.
    * <p>
    * This attribute is kept as a user-data at the DOM Node, to be shared by all resources 
    * referencing this SOAP Header Entry.
    * <p>
    * The {@code mustUnderstand} and {@code actor} attributes are not emitted by 
    * {@link #getSerialized()}.
    *
    * @param    uri
    *           The URI of the SOAP actor.
    * 
    * @return   <code>true</code> if the operation was possible.
    */
   public logical setActor(String uri)
   {
      return setActor(new character(uri));
   }

   /**
    * Sets the MUST-UNDERSTAND attribute for this SOAP-HEADER-ENTRY.
    * <p>
    * This attribute is kept as a user-data at the DOM Node, to be shared by all resources 
    * referencing this SOAP Header Entry.
    * <p>
    * The {@code mustUnderstand} and {@code actor} attributes are not emitted by 
    * {@link #getSerialized()}.
    * 
    * @param    val
    *           The new value of this attribute.
    * 
    * @return   <code>true</code> if the operation was possible.
    */
   public logical setMustUnderstand(logical val)
   {
      Node node = getHeaderNode();

      if (!isPopulated("SET-MUST-UNDERSTAND") || node == null)
      {
         return new logical(false);
      }

      node.setUserData(MUST_UNDERSTAND, val.booleanValue(), null);

      return new logical(true);
   }

   /**
    * Sets the MUST-UNDERSTAND attribute for this SOAP-HEADER-ENTRY.
    *
    * @param    val
    *           The new value of this attribute.
    * 
    * @return   <code>true</code> if the operation was possible.
    */
   public logical setMustUnderstand(boolean val)
   {
      return setMustUnderstand(new logical(val));
   }

   /**
    * Replaces the underlying XML of a SOAP-HEADER-ENTRYREF object with the DOM XML sub-tree 
    * rooted by the X-NODEREF object.
    *
    * @param    xNoderef
    *           A X-NODEREF variable that refers to the XML that will be set for the 
    *           SOAP-HEADER-ENTRYREF object.
    * 
    * @return   <code>true</code> if the operation was possible.
    */
   public logical setNode(handle xNoderef)
   {
      if (xNoderef._isValid() && (xNoderef.getResource() instanceof XNodeRefImpl))
      {
         XNodeRefImpl nodeRef = (XNodeRefImpl) xNoderef.getResource();
         
         // check if there is a valid created DOM node
         Node domNode = nodeRef.getNode();
         
         if (domNode == null)
         {
            final String msg1 = "Invalid X-NODEREF handle passed to SET-NODE";
            final int code1 = 11577;

            final String msg2 = 
              "**The SET-NODE attribute on the SOAP-HEADER-ENTRYREF widget has invalid arguments";
            final int code2 = 4065;

            ErrorManager.recordOrShowError(new int[] {code1, code2}, new String[]{msg1, msg2},
                                           false, false, false);
            
            return new logical(false);
         }
         
         // check if it is attached to a SOAP Header
         if (parent == null)
         {
            final String msg = "SET-NODE called on an unattached SOAP-HEADER-ENTRYREF";
            ErrorManager.recordOrShowError(13877, msg, false, false, false);
            
            return new logical(false);
         }
         
         // create a deep clone of the target node
         parent.setHeaderNode(headerKey, domNode.cloneNode(true));
         
         return new logical(true);
      }

      // no error is shown
      
      return new logical(false);
   }

   /**
    * Sets the SOAP-header-entryref object's underlying XML from serialized XML.
    *
    * @param    payload
    *           A LONGCHAR variable that contains the XML that will become this 
    *           SOAP-HEADER-ENTRYREF object. This parameter can be only a longchar, 4GL doesn't
    *           allow character variables.
    * 
    * @return   <code>true</code> if the operation was possible.
    */
   public logical setSerialized(longchar payload)
   {
      // check if it is attached to a SOAP Header
      if (parent == null)
      {
         final String msg = "SET-SERIALIZED called on an unattached SOAP-HEADER-ENTRYREF";
         ErrorManager.recordOrShowError(13877, msg, false, false, false);
         
         return new logical(false);
      }

      String xml = payload.toStringMessage();

      try
      {
         Document doc = XmlHelper.parse(new ByteArrayInputStream(xml.getBytes()),
                                        false, null, true);
         
         parent.setHeaderNode(headerKey, doc.getDocumentElement());
         doc.setUserData(XNodeRefImpl.RESTRICTED_DOCUMENT, true, null);
         
         return new logical(true);
      }
      catch (ParserConfigurationException | IOException | SAXException e)
      {
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, "Problems parsing the XML payload: " + xml, e);
         }
         
         final String msg1 = "Error in SET-SERIALIZED:";
         final String msg2 =
            "**The SET-SERIALIZED attribute on the SOAP-HEADER-ENTRYREF widget has invalid " +
            "arguments. (4065)";
         
         ErrorManager.recordOrShowError(new int[] {0, 4065}, new String[] {msg1, msg2},
                                        false, false, false, true);
         
         return new logical(false);
      }
   }
   
   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return !deleted;
   }

   /**
    * Get the value of the LOCAL-NAME attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getLocalName()
   {
      Node node = getHeaderNode();
      
      if (node == null)
      {
         return new character("");
      }

      String name = node.getLocalName();
      return new character(name == null ? "" : name);
   }

   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return new character(this.admData);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      this.admData.assign(value);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      this.admData.assign(value);
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself.
    * 
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      return new integer(this.uniqueID);
   }

   /**
    * Returns the value of the the NAMESPACE-URI attribute.
    *
    * @return    See above.
    */
   @Override
   public character getNamespaceURI()
   {
      Node node = getHeaderNode();
      
      if (node == null)
      {
         return new character("");
      }
      
      String uri = node.getNamespaceURI();
      return new character(uri == null ? "" : uri);
   }

   /**
    * Setter for NAMESPACE-URI attribute. Calls {@link #readOnlyError}, as this attribute is r/o
    * for this resource.
    *
    * @param    uri
    *           Not used.
    */
   @Override
   public void setNamespaceURI(String uri)
   {
      readOnlyError("NAMESPACE-URI");
   }

   /**
    * Setter for NAMESPACE-URI attribute. Calls {@link #readOnlyError}, as this attribute is r/o
    * for this resource.
    *
    * @param    uri
    *           Not used.
    */
   @Override
   public void setNamespaceURI(character uri)
   {
      readOnlyError("NAMESPACE-URI");
   }

   /**
    * Get the <code>name</code> attribute of handle.
    * 
    * @return   See above.
    */
   public character name()
   {
      return getLocalName();
   }

   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      // the inner X-NODEREF does not get deleted. just mark the resource.

      this.deleted = true;

      return true;
   }
   
   /**
    * Check if this resource supports the NEXT-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasNextSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the PREV-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasPrevSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the NAME attribute in read-only mode.
    * 
    * @return   Always <code>true</code>.
    */
   @Override
   protected boolean hasNameReadOnly()
   {
      return true;
   }

   /**
    * Set the parent SOAP HEADER for this entry.
    * 
    * @param    parent
    *           The parent SOAP Header.
    * @param    headerKey
    *           The key to the DOM XML Node referenced by this header entry.
    */
   void setParent(SOAPHeaderImpl parent, Object headerKey)
   {
      this.parent = parent;
      this.headerKey = headerKey;
   }
   
   /**
    * Get the {@link #headerKey} for this SOAP header entry.
    * 
    * @return   See above.
    */
   Object getHeaderKey()
   {
      return headerKey;
   }
   
   /**
    * Check if this SOAP-HEADER-ENTRY is populated. If not, show an appropriate error.
    * 
    * @param    caller
    *           The name of the caller of this method (will be shown in the error).
    *           
    * @return   See above.
    */
   private boolean isPopulated(String caller)
   {
      Node node = getHeaderNode();
      
      if (node == null)
      {
         final String msg = "%s called on an unpopulated SOAP-HEADER-ENTRYREF";
         
         ErrorManager.recordOrShowError(11571, String.format(msg, caller), false, false, false);
      }
      
      return node != null;
   }
   
   /**
    * Get the {@link Node DOM XML node} associated with this SOAP Header entry.
    * 
    * @return   See above.
    */
   private Node getHeaderNode()
   {
      return parent == null ? null : parent.getHeaderNode(headerKey);
   }
}
