/*
** Module   : ServerHelper.java
** Abstract : defines common APIs for servers connected to an appserver or web service.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 CA  20131220 Created initial version.
*/

package com.goldencode.p2j.util;

/**
 * Defines common APIs used by a server connected to either an appserver or a web service.
 */
public abstract class ServerHelper
{
   /** The {@link ServerImpl} instance associated with this helper. */
   protected ServerImpl server;

   /**
    * Get the name of this server.
    * 
    * @return   See above.
    */
   public abstract String getServerName();

   /**
    * Get the sub-type of this server.
    * 
    * @return   See above.
    */
   public abstract String getSubType();

   /**
    * Get the connection ID.
    * 
    * @return   See above.
    */
   public abstract String getConnectionID();
   
   /**
    * Check if this is a web service.
    * 
    * @return   <code>true</code> if we are connected to a web service.
    */
   public abstract boolean isWebService();
   
   /**
    * Check if this is a session-free connection.
    * 
    * @return   <code>true</code> if this is a session-free connection.
    */
   public abstract boolean isSessionFree();
   
   /**
    * Disconnect this server.
    */
   public abstract void disconnect();
   
   /**
    * Send a STOP condition to the specified async request.
    * 
    * @param    requestId
    *           The request where to send a STOP condition.
    */
   public abstract void sendStop(int requestId);

   /**
    * Invoke a remote external procedure, internal procedure, user-defined function or web service
    * operation. The validation for the name and arguments will be specific to the established 
    * server connection subtype.
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    h
    *           The persistent procedure in which to invoke the internal procedure, function or
    *           operation.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of eac parameter. The modes may be mandatory 
    *           for a specific implementation.
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this procedure/function invocation or <code>null</code> if the
    *           invocation could not be performed.
    */
   abstract BaseDataType invoke(AsyncRequestImpl asyncReq,
                                character        name,
                                handle           h,
                                boolean          function,
                                boolean          dynamicFunction,
                                boolean          superCall,
                                boolean          transactionDistinct,
                                String           modes,
                                Object...        args);
   
   /**
    * Set the {@link ServerImpl} instance associated with this helper.
    * 
    * @param    server
    *           The {@link ServerImpl} instance.
    */
   void setServer(ServerImpl server)
   {
      this.server = server;
   }

   /**
    * Get the {@link #server} instance associated with this helper.
    * 
    * @return   See above.
    */
   ServerImpl getServer()
   {
      return server;
   }
}
